# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from karborclient.osc.v1 import operation_logs as osc_operation_logs
from karborclient.tests.unit.osc.v1 import fakes
from karborclient.v1 import operation_logs


OPERATIONLOG_INFO = {
    "id": "22b82aa7-9179-4c71-bba2-caf5c0e68db7",
    "project_id": "e486a2f49695423ca9c47e589b948108",
    "operation_type": "protect",
    "checkpoint_id": "dcb20606-ad71-40a3-80e4-ef0fafdad0c3",
    "plan_id": "cf56bd3e-97a7-4078-b6d5-f36246333fd9",
    "provider_id": "23902b02-5666-4ee6-8dfe-962ac09c3994",
    "restore_id": None,
    "scheduled_operation_id": "23902b02-5666-4ee6-8dfe-962ac09c3991",
    "started_at": "2015-08-27T09:50:58-05:00",
    "ended_at": "2015-08-27T10:50:58-05:00",
    "status": "protecting",
    "error_info": "Could not access bank",
    "extra_info": None
}


class TestOperationLogs(fakes.TestDataProtection):
    def setUp(self):
        super(TestOperationLogs, self).setUp()
        self.operation_logs_mock = (
            self.app.client_manager.data_protection.operation_logs)
        self.operation_logs_mock.reset_mock()


class TestListOperationLogs(TestOperationLogs):
    def setUp(self):
        super(TestListOperationLogs, self).setUp()
        self.operation_logs_mock.list.return_value = [
            operation_logs.OperationLog(None,
                                        copy.deepcopy(OPERATIONLOG_INFO))]

        # Command to test
        self.cmd = osc_operation_logs.ListOperationLogs(self.app, None)

    def test_operation_logs_list(self):
        arglist = ['--status', 'success']
        verifylist = [('status', 'success')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = (
            ['Id', 'Operation Type', 'Checkpoint id', 'Plan Id',
             'Provider id', 'Restore Id', 'Scheduled Operation Id',
             'Status', 'Started At', 'Ended At', 'Error Info',
             'Extra Info'])
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [("22b82aa7-9179-4c71-bba2-caf5c0e68db7",
                          "protect",
                          "dcb20606-ad71-40a3-80e4-ef0fafdad0c3",
                          "cf56bd3e-97a7-4078-b6d5-f36246333fd9",
                          "23902b02-5666-4ee6-8dfe-962ac09c3994",
                          None,
                          "23902b02-5666-4ee6-8dfe-962ac09c3991",
                          "protecting",
                          "2015-08-27T09:50:58-05:00",
                          "2015-08-27T10:50:58-05:00",
                          "Could not access bank",
                          None)]
        self.assertEqual(expected_data, list(data))


class TestShowOperationLog(TestOperationLogs):
    def setUp(self):
        super(TestShowOperationLog, self).setUp()
        self._oplog_info = copy.deepcopy(OPERATIONLOG_INFO)
        self.operation_logs_mock.get.return_value = (
            operation_logs.OperationLog(None, self._oplog_info))

        # Command to test
        self.cmd = osc_operation_logs.ShowOperationLog(self.app, None)

    def test_operation_log_show(self):
        arglist = ['22b82aa7-9179-4c71-bba2-caf5c0e68db7']
        verifylist = [('operation_log',
                       '22b82aa7-9179-4c71-bba2-caf5c0e68db7')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = (
            'checkpoint_id', 'ended_at', 'error_info', 'extra_info',
            'id', 'operation_type', 'plan_id', 'project_id',
            'provider_id', 'restore_id', 'scheduled_operation_id',
            'started_at', 'status')
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        self.assertEqual(self._oplog_info['checkpoint_id'], data[0])
        self.assertEqual(self._oplog_info['ended_at'], data[1])
        self.assertEqual(self._oplog_info['error_info'], data[2])
        self.assertEqual(self._oplog_info['extra_info'], data[3])
        self.assertEqual(self._oplog_info['id'], data[4])
        self.assertEqual(self._oplog_info['operation_type'], data[5])
        self.assertEqual(self._oplog_info['plan_id'], data[6])
        self.assertEqual(self._oplog_info['project_id'], data[7])
        self.assertEqual(self._oplog_info['provider_id'], data[8])
        self.assertEqual(self._oplog_info['restore_id'], data[9])
        self.assertEqual(self._oplog_info['scheduled_operation_id'], data[10])
        self.assertEqual(self._oplog_info['started_at'], data[11])
        self.assertEqual(self._oplog_info['status'], data[12])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from karborclient.common.apiclient import exceptions
from karborclient.tests.unit import base
from karborclient.tests.unit.v1 import fakes

cs = fakes.FakeClient()
mock_request_return = ({}, {'trigger_info': {'name': 'fake_name'}})


class TriggersTest(base.TestCaseShell):

    @mock.patch('karborclient.common.http.HTTPClient.json_request')
    def test_list_triggers(self, mock_request):
        mock_request.return_value = mock_request_return
        cs.triggers.list()
        mock_request.assert_called_with(
            'GET',
            '/triggers', headers={})

    @mock.patch('karborclient.common.http.HTTPClient.json_request')
    def test_list_triggers_with_all_tenants(self, mock_request):
        mock_request.return_value = mock_request_return
        cs.triggers.list(search_opts={'all_tenants': 1})
        mock_request.assert_called_with(
            'GET',
            '/triggers?all_tenants=1', headers={})

    @mock.patch('karborclient.common.http.HTTPClient.json_request')
    def test_list_triggers_with_marker_limit(self, mock_request):
        mock_request.return_value = mock_request_return
        cs.triggers.list(marker=1234, limit=2)
        mock_request.assert_called_with(
            'GET',
            '/triggers?limit=2&marker=1234', headers={})

    @mock.patch('karborclient.common.http.HTTPClient.json_request')
    def test_list_triggers_with_sort_key_dir(self, mock_request):
        mock_request.return_value = mock_request_return
        cs.triggers.list(sort_key='id', sort_dir='asc')
        mock_request.assert_called_with(
            'GET',
            '/triggers?'
            'sort_dir=asc&sort_key=id', headers={})

    @mock.patch('karborclient.common.http.HTTPClient.json_request')
    def test_list_triggers_with_invalid_sort_key(self, mock_request):
        self.assertRaises(ValueError,
                          cs.triggers.list, sort_key='invalid', sort_dir='asc')

    @mock.patch('karborclient.common.http.HTTPClient.json_request')
    def test_create_trigger(self, mock_request):
        mock_request.return_value = mock_request_return
        cs.triggers.create('name', 'time', {})
        mock_request.assert_called_with(
            'POST',
            '/triggers',
            data={
                'trigger_info': {'name': 'name',
                                 'type': 'time',
                                 'properties': {}}},
            headers={})

    def test_create_trigger_with_invalid_window(self):
        self.assertRaises(exceptions.CommandError,
                          cs.triggers.create,
                          'name', 'time', {'window': 'fake'})

    @mock.patch('karborclient.common.http.HTTPClient.raw_request')
    def test_delete_trigger(self, mock_request):
        mock_request.return_value = mock_request_return
        cs.triggers.delete('1')
        mock_request.assert_called_with(
            'DELETE',
            '/triggers/1',
            headers={})

    @mock.patch('karborclient.common.http.HTTPClient.json_request')
    def test_show_trigger(self, mock_request):
        mock_request.return_value = mock_request_return
        cs.triggers.get('1')
        mock_request.assert_called_with(
            'GET',
            '/triggers/1',
            headers={})

    @mock.patch('karborclient.common.http.HTTPClient.json_request')
    def test_show_trigger_with_headers(self, mock_request):
        mock_request.return_value = mock_request_return
        cs.triggers.get('1', session_id='fake_session_id')
        mock_request.assert_called_with(
            'GET',
            '/triggers/1',
            headers={'X-Configuration-Session': 'fake_session_id'})

    @mock.patch('karborclient.common.http.HTTPClient.json_request')
    def test_update_trigger(self, mock_request):
        mock_request.return_value = mock_request_return
        trigger_id = '123'
        data = {"name": "My Trigger",
                "properties": {"pattern": "0 10 * * *", "format": "crontab"}}
        body = {"trigger_info": data}
        cs.triggers.update(trigger_id, data)
        mock_request.assert_called_with(
            'PUT',
            '/triggers/123',
            data=body,
            headers={}
        )

    def test_update_trigger_with_invalid_window(self):
        trigger_id = '123'
        self.assertRaises(exceptions.CommandError,
                          cs.triggers.update,
                          trigger_id, {'properties': {'window': 'fake'}})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import mock

from karborclient import shell
from karborclient.tests.unit import base
from karborclient.tests.unit.v1 import fakes

FAKE_PROVIDER_ID = '1234'
FAKE_ENDPOINT = 'http://127.0.0.1/identity'


class ShellFixture(fixtures.Fixture):
    def setUp(self):
        super(ShellFixture, self).setUp()
        self.shell = shell.KarborShell()

    def tearDown(self):
        # For some method like test_image_meta_bad_action we are
        # testing a SystemExit to be thrown and object self.shell has
        # no time to get instantiated which is OK in this case, so
        # we make sure the method is there before launching it.
        if hasattr(self.shell, 'cs'):
            self.shell.cs.clear_callstack()
        super(ShellFixture, self).tearDown()


class ShellTest(base.TestCaseShell):

    FAKE_ENV = {
        'OS_USERNAME': 'username',
        'OS_PASSWORD': 'password',
        'OS_TENANT_NAME': 'project_id',
        'OS_AUTH_URL': 'http://no.where/v2.0',
        'OS_AUTH_TOKEN': 'fake_token'
    }

    def setUp(self):
        """Run before each test."""
        super(ShellTest, self).setUp()
        for var in self.FAKE_ENV:
            self.useFixture(fixtures.EnvironmentVariable(
                var, self.FAKE_ENV[var]))
        self.shell = self.useFixture(ShellFixture()).shell

        get_endpoint = mock.MagicMock()
        get_endpoint.return_value = FAKE_ENDPOINT
        self.useFixture(fixtures.MonkeyPatch(
            'keystoneauth1.identity.generic.token.Token.get_endpoint',
            get_endpoint))
        self.useFixture(fixtures.MonkeyPatch('karborclient.client.Client',
                                             fakes.FakeClient))
        self.useFixture(fixtures.MonkeyPatch(
            'karborclient.common.http._construct_http_client',
            fakes.FakeHTTPClient))

    def run_command(self, cmd):
        if not isinstance(cmd, list):
            cmd = cmd.split()
        self.shell.main(cmd)

    def assert_called(self, method, url, body=None, **kwargs):
        return self.shell.cs.assert_called(method, url, body, **kwargs)

    def test_checkpoint_list_with_all_tenants(self):
        self.run_command(
            'checkpoint-list ' + FAKE_PROVIDER_ID + ' --all-tenants 1')

        self.assert_called('GET',
                           '/providers/1234/'
                           'checkpoints?all_tenants=1')

    def test_checkpoint_list_with_all(self):
        self.run_command(
            'checkpoint-list ' + FAKE_PROVIDER_ID + ' --all')
        self.assert_called('GET',
                           '/providers/1234/'
                           'checkpoints?all_tenants=1')

    def test_plan_list_with_all_tenants(self):
        self.run_command('plan-list --all-tenants 1')
        self.assert_called('GET', '/plans?all_tenants=1')

    def test_plan_list_with_all(self):
        self.run_command('plan-list --all')
        self.assert_called('GET', '/plans?all_tenants=1')

    def test_resotre_list_with_all_tenants(self):
        self.run_command('restore-list --all-tenants 1')
        self.assert_called('GET', '/restores?all_tenants=1')

    def test_resotre_list_with_all(self):
        self.run_command('restore-list --all')
        self.assert_called('GET', '/restores?all_tenants=1')

    def test_verification_list_with_all_tenants(self):
        self.run_command('verification-list --all-tenants 1')
        self.assert_called('GET', '/verifications?all_tenants=1')

    def test_verification_list_with_all(self):
        self.run_command('verification-list --all')
        self.assert_called('GET', '/verifications?all_tenants=1')

    def test_trigger_list_with_all_tenants(self):
        self.run_command('trigger-list --all-tenants 1')
        self.assert_called('GET', '/triggers?all_tenants=1')

    def test_trigger_list_with_all(self):
        self.run_command('trigger-list --all')
        self.assert_called('GET', '/triggers?all_tenants=1')

    def test_scheduledoperation_list_with_all_tenants(self):
        self.run_command('scheduledoperation-list --all-tenants 1')
        self.assert_called('GET', '/scheduled_operations?all_tenants=1')

    def test_scheduledoperation_list_with_all(self):
        self.run_command('scheduledoperation-list --all')
        self.assert_called('GET', '/scheduled_operations?all_tenants=1')

    def test_operationlog_list_with_all_tenants(self):
        self.run_command('operationlog-list --all-tenants 1')
        self.assert_called('GET', '/operation_logs?all_tenants=1')

    def test_operationlog_list_with_all(self):
        self.run_command('operationlog-list --all')
        self.assert_called('GET', '/operation_logs?all_tenants=1')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Data protection V1 plan action implementations"""

from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from osc_lib.command import command
from osc_lib import utils as osc_utils
from oslo_log import log as logging

from karborclient.common.apiclient import exceptions
from karborclient.i18n import _
from karborclient import utils


def format_plan(plan_info):
    for key in ('resources', 'parameters'):
        if key not in plan_info:
            continue
        plan_info[key] = jsonutils.dumps(plan_info[key],
                                         indent=2, sort_keys=True)
    plan_info.pop("links", None)


class ListPlans(command.Lister):
    _description = _("List plans.")

    log = logging.getLogger(__name__ + ".ListPlans")

    def get_parser(self, prog_name):
        parser = super(ListPlans, self).get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=_('Include all projects (admin only)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Filter results by plan name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Filter results by plan description'),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            help=_('Filter results by status'),
        )
        parser.add_argument(
            '--marker',
            metavar='<plan>',
            help=_('The last plan ID of the previous page'),
        )
        parser.add_argument(
            '--limit',
            type=int,
            metavar='<num-plans>',
            help=_('Maximum number of plans to display'),
        )
        parser.add_argument(
            '--sort',
            metavar="<key>[:<direction>]",
            default=None,
            help=_("Sort output by selected keys and directions(asc or desc) "
                   "(default: name:asc), multiple keys and directions can be "
                   "specified separated by comma"),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Filter results by a project(admin only)')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        data_protection_client = self.app.client_manager.data_protection
        all_projects = bool(parsed_args.project) or parsed_args.all_projects

        search_opts = {
            'all_tenants': all_projects,
            'project_id': parsed_args.project,
            'name': parsed_args.name,
            'description': parsed_args.description,
            'status': parsed_args.status,
        }

        data = data_protection_client.plans.list(
            search_opts=search_opts, marker=parsed_args.marker,
            limit=parsed_args.limit, sort=parsed_args.sort)

        column_headers = ['Id', 'Name', 'Description', 'Provider id', 'Status']

        return (column_headers,
                (osc_utils.get_item_properties(
                    s, column_headers
                ) for s in data))


class ShowPlan(command.ShowOne):
    _description = "Shows plan details"

    def get_parser(self, prog_name):
        parser = super(ShowPlan, self).get_parser(prog_name)
        parser.add_argument(
            'plan',
            metavar="<plan>",
            help=_('The UUID of the plan.')
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.data_protection
        plan = osc_utils.find_resource(client.plans, parsed_args.plan)

        format_plan(plan._info)
        return zip(*sorted(plan._info.items()))


class CreatePlan(command.ShowOne):
    _description = "Creates a plan"

    def get_parser(self, prog_name):
        parser = super(CreatePlan, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_('The name of the plan.')
        )
        parser.add_argument(
            'provider_id',
            metavar='<provider_id>',
            help=_('The UUID of the provider.')
        )
        parser.add_argument(
            'resources',
            metavar='<id=type=name=extra_info,id=type=name=extra_info>',
            help=_('Resource in list must be a dict when creating'
                   ' a plan. The keys of resource are id ,type, name and '
                   'extra_info. The extra_info field is optional.')
        )
        parser.add_argument(
            '--parameters-json',
            type=str,
            dest='parameters_json',
            metavar='<parameters>',
            default=None,
            help=_('Plan parameters in json format.')
        )
        parser.add_argument(
            '--parameters',
            action='append',
            metavar='resource_type=<type>[,resource_id=<id>,key=val,...]',
            default=[],
            help=_('Plan parameters, may be specified multiple times. '
                   'resource_type: type of resource to apply parameters. '
                   'resource_id: limit the parameters to a specific resource. '
                   'Other keys and values: according to provider\'s protect '
                   'schema.')
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('The description of the plan.')
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.data_protection
        if not uuidutils.is_uuid_like(parsed_args.provider_id):
            raise exceptions.CommandError(
                "Invalid provider id provided.")
        plan_resources = utils.extract_resources(parsed_args)
        utils.check_resources(client, plan_resources)
        plan_parameters = utils.extract_parameters(parsed_args)
        plan = client.plans.create(parsed_args.name, parsed_args.provider_id,
                                   plan_resources, plan_parameters,
                                   description=parsed_args.description)

        format_plan(plan._info)
        return zip(*sorted(plan._info.items()))


class UpdatePlan(command.ShowOne):
    _description = "Update a plan"

    def get_parser(self, prog_name):
        parser = super(UpdatePlan, self).get_parser(prog_name)
        parser.add_argument(
            "plan_id",
            metavar="<PLAN ID>",
            help=_("Id of plan to update.")
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            help=_("A name to which the plan will be renamed.")
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Description to which the plan will be updated.")
        )
        parser.add_argument(
            "--resources",
            metavar="<id=type=name,id=type=name>",
            help=_("Resources to which the plan will be updated.")
        )
        parser.add_argument(
            "--status",
            metavar="<suspended|started>",
            help=_("status to which the plan will be updated.")
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.data_protection
        data = {}
        if parsed_args.name is not None:
            data['name'] = parsed_args.name
        if parsed_args.description is not None:
            data['description'] = parsed_args.description
        if parsed_args.resources is not None:
            plan_resources = utils.extract_resources(parsed_args)
            data['resources'] = plan_resources
        if parsed_args.status is not None:
            data['status'] = parsed_args.status
        try:
            plan = osc_utils.find_resource(client.plans,
                                           parsed_args.plan_id)
            plan = client.plans.update(plan.id, data)
        except exceptions.NotFound:
            raise exceptions.CommandError(
                "Plan %s not found" % parsed_args.plan_id)
        else:
            format_plan(plan._info)
            return zip(*sorted(plan._info.items()))


class DeletePlan(command.Command):
    _description = "Delete plan"

    def get_parser(self, prog_name):
        parser = super(DeletePlan, self).get_parser(prog_name)
        parser.add_argument(
            'plan',
            metavar='<plan>',
            nargs="+",
            help=_('ID of plan.')
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.data_protection
        failure_count = 0
        for plan_id in parsed_args.plan:
            try:
                plan = osc_utils.find_resource(client.plans, plan_id)
                client.plans.delete(plan.id)
            except exceptions.NotFound:
                failure_count += 1
                print("Failed to delete '{0}'; plan not "
                      "found".format(plan_id))
        if failure_count == len(parsed_args.plan):
            raise exceptions.CommandError(
                "Unable to find and delete any of the "
                "specified plan.")
